# Copyright 2018  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=



import logging
import os
import tempfile

import cliapp

import vmdb


class LuksPlugin(cliapp.Plugin):

    def enable(self):
        self.app.step_runners.add(CryptsetupStepRunner())


class CryptsetupStepRunner(vmdb.StepRunnerInterface):

    def get_required_keys(self):
        return ['cryptsetup']

    def run(self, step, settings, state):
        underlying = step['cryptsetup']
        crypt_name = step['tag']

        state.tmp_key_file = None
        key_file = step.get('key-file')
        key_cmd = step.get('key-cmd')
        if key_file is None and key_cmd is None:
            raise Exception(
                'cryptsetup step MUST define one of key-file or key-cmd')

        if key_file is None:
            output = vmdb.runcmd(['sh', '-c', key_cmd])
            output = output.decode('UTF-8')
            key = output.splitlines()[0]
            fd, key_file = tempfile.mkstemp()
            state.tmp_key_file = key_file
            os.close(fd)
            open(key_file, 'w').write(key)

        dev = state.tags.get_dev(underlying)
        if dev is None:
            for t in state.tags.get_tags():
                logging.debug(
                    'tag %r dev %r mp %r', t, state.tags.get_dev(t),
                    state.tags.get_mount_point(t))
            assert 0

        vmdb.runcmd(['cryptsetup', 'luksFormat', dev, key_file])
        vmdb.runcmd(
            ['cryptsetup', 'open', '--type', 'luks', '--key-file', key_file, dev,
             crypt_name])

        crypt_dev = '/dev/mapper/{}'.format(crypt_name)
        assert os.path.exists(crypt_dev)
        state.tags.append(crypt_name)
        state.tags.set_dev(crypt_name, crypt_dev)

    def teardown(self, step, settings, state):
        x = state.tmp_key_file
        if x is not None and os.path.exists(x):
            os.remove(x)

        crypt_name = step['tag']

        crypt_dev = '/dev/mapper/{}'.format(crypt_name)
        vmdb.runcmd(['cryptsetup', 'close', crypt_dev])



# Copyright 2018  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=



import os

import cliapp

import vmdb


class Lvm2Plugin(cliapp.Plugin):

    def enable(self):
        self.app.step_runners.add(VgcreateStepRunner())
        self.app.step_runners.add(LvcreateStepRunner())


class VgcreateStepRunner(vmdb.StepRunnerInterface):

    def get_required_keys(self):
        return ['vgcreate', 'physical']

    def run(self, step, settings, state):
        vgname = self.get_vg(step)
        physical = self.get_pv(step, state)

        for phys in physical:
            vmdb.runcmd(['pvcreate', '-ff', '--yes', phys])
        vmdb.runcmd(['vgcreate', vgname] + physical)

    def teardown(self, step, settings, state):
        vgname = self.get_vg(step)
        vmdb.runcmd(['vgchange', '-an', vgname])

    def get_vg(self, step):
        return step['vgcreate']

    def get_pv(self, step, state):
        return [
            state.tags.get_dev(tag)
            for tag in step['physical']
        ]


class LvcreateStepRunner(vmdb.StepRunnerInterface):

    def get_required_keys(self):
        return ['lvcreate']

    def run(self, step, settings, state):
        vgname = step['lvcreate']
        lvname = step['name']
        size = step['size']

        vmdb.runcmd(['lvcreate', '--name', lvname, '--size', size, vgname])

        lvdev = '/dev/{}/{}'.format(vgname, lvname)
        assert os.path.exists(lvdev)
        state.tags.append(lvname)
        state.tags.set_dev(lvname, lvdev)



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=



import cliapp

import vmdb


class DebootstrapPlugin(cliapp.Plugin):

    def enable(self):
        self.app.step_runners.add(DebootstrapStepRunner())


class DebootstrapStepRunner(vmdb.StepRunnerInterface):

    def get_required_keys(self):
        return ['debootstrap', 'target', 'mirror']

    def run(self, step, settings, state):
        suite = step['debootstrap']
        tag = step['target']
        target = state.tags.get_mount_point(tag)
        mirror = step['mirror']
        variant = step.get('variant', '-')
        if not (suite and tag and target and mirror):
            raise Exception('missing arg for debootstrap step')
        vmdb.runcmd(['debootstrap', '--variant', variant, suite, target, mirror])

    def run_even_if_skipped(self, step, settings, state):
        tag = step['target']
        target = state.tags.get_mount_point(tag)
        vmdb.runcmd_chroot(target, ['apt-get', 'update'])



# Copyright 2018  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


# Unmount a directory, including any mount points under that
# directory. If /mnt/foo is given, and /mnt/foo/bar is also mounted,
# unmount /mnt/foo/bar first, and /mnt/foo then. Look for sub-mounts
# in /proc/mounts.


class Tags:

    def __init__(self):
        self._tags = {}
        self._tagnames = []

    def get_tags(self):
        return list(self._tags.keys())

    def has_tag(self, tag):
        return tag in self._tags

    def get_dev(self, tag):
        item = self._get(tag)
        return item['dev']

    def get_mount_point(self, tag):
        item = self._get(tag)
        return item['mount_point']

    def is_cached(self, tag):
        item = self._get(tag)
        return item.get('cached', False)

    def append(self, tag):
        if tag in self._tags:
            raise TagInUse(tag)
        self._tagnames.append(tag)
        self._tags[tag] = {
            'dev': None,
            'mount_point': None,
        }

    def set_dev(self, tag, dev):
        item = self._get(tag)
        if item['dev'] is not None:
            raise AlreadyHasDev(tag)
        item['dev'] = dev

    def set_mount_point(self, tag, mount_point, cached=False):
        item = self._get(tag)
        if item['mount_point'] is not None:
            raise AlreadyMounted(tag)
        item['mount_point'] = mount_point
        item['cached'] = cached

    def _get(self, tag):
        item = self._tags.get(tag)
        if item is None:
            raise UnknownTag(tag)
        return item


class TagInUse(Exception):

    def __init__(self, tag):
        super().__init__('Tag already used: {}'.format(tag))


class UnknownTag(Exception):

    def __init__(self, tag):
        super().__init__('Unknown tag: {}'.format(tag))


class AlreadyHasDev(Exception):

    def __init__(self, tag):
        super().__init__('Already has device: {}'.format(tag))


class AlreadyMounted(Exception):

    def __init__(self, tag):
        super().__init__('Already mounted tag: {}'.format(tag))



# Copyright 2019 Gunnar Wolf
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=

import cliapp
import vmdb
import os
import logging

class CreateFilePlugin(cliapp.Plugin):
    def enable(self):
        self.app.step_runners.add(CreateFileStepRunner())
        self.app.step_runners.add(CreateDirStepRunner())

class CreateFileStepRunner(vmdb.StepRunnerInterface):
    def get_required_keys(self):
        return ['create-file', 'contents']

    def run(self, step, settings, state):
        root = state.tags.get_mount_point('/')
        newfile = step['create-file']
        contents = step['contents']
        perm = step.get('perm')
        uid = step.get('uid')
        gid = step.get('gid')

        filename = '/'.join([root,newfile])

        if perm:
            perm = int(perm, 8)
        else:
            perm = 0o0644
        if uid:
            uid = int(uid)
        else:
            uid = 0
        if gid:
            gid = int(gid)
        else:
            gid = 0

        logging.info('Creating file %s, uid %d, gid %d, perms %o' % (filename, uid, gid, perm))
        fd = open(filename, 'w')
        fd.write(contents)
        fd.close

        os.chown(filename, uid, gid)
        os.chmod(filename, perm)

class CreateDirStepRunner(vmdb.StepRunnerInterface):
    def get_required_keys(self):
        return ['create-dir']

    def run(self, step, settings, state):
        root = state.tags.get_mount_point('/')
        newdir = step['create-dir']
        path = '/'.join([root, newdir])
        perm = step.get('perm')
        uid = step.get('uid')
        gid = step.get('gid')

        if perm:
            perm = int(perm, 8)
        else:
            perm = 0o0755
        if uid:
            uid = int(uid)
        else:
            uid = 0
        if gid:
            gid = int(gid)
        else:
            gid = 0

        logging.info('Creating directory %s, uid %d, gid %d, perms %o' % (path, uid, gid, perm))

        os.makedirs(path, perm)
        os.chown(path, uid, gid)



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=



import os
import stat

import cliapp

import vmdb


class PartitionPlugin(cliapp.Plugin):

    def enable(self):
        self.app.step_runners.add(MklabelStepRunner())
        self.app.step_runners.add(MkpartStepRunner())
        self.app.step_runners.add(KpartxStepRunner())


class MklabelStepRunner(vmdb.StepRunnerInterface):

    def get_required_keys(self):
        return ['mklabel', 'device']

    def run(self, step, settings, state):
        label_type = step['mklabel']
        device = step['device']
        vmdb.runcmd(['parted', '-s', device, 'mklabel', label_type])


class MkpartStepRunner(vmdb.StepRunnerInterface):

    def get_required_keys(self):
        return ['mkpart', 'device', 'start', 'end']

    def run(self, step, settings, state):
        part_type = step['mkpart']
        device = step['device']
        start = step['start']
        end = step['end']
        tag = step.get('tag') or step.get('part-tag')
        if tag is None:
            tag = step['part-tag']
        fs_type = step.get('fs-type', 'ext2')

        orig = self.list_partitions(device)
        vmdb.runcmd(['parted', '-s', device, 'mkpart', part_type, fs_type, start, end])

        state.tags.append(tag)
        if self.is_block_dev(device):
            new = self.list_partitions(device)
            diff = self.diff_partitions(orig, new)
            assert len(diff) == 1
            vmdb.progress('remembering partition {} as {}'.format(diff[0], tag))
            state.tags.set_dev(tag, diff[0])

    def is_block_dev(self, filename):
        st = os.lstat(filename)
        return stat.S_ISBLK(st.st_mode)

    def list_partitions(self, device):
        output = vmdb.runcmd(['parted', '-m', device, 'print'])
        output = output.decode('UTF-8')
        partitions = [
            line.split(':')[0]
            for line in output.splitlines()
            if ':' in line
        ]
        return [
            word if word.startswith('/') else '{}{}'.format(device, word)
            for word in partitions
        ]

    def diff_partitions(self, old, new):
        return [
            line
            for line in new
            if line not in old
        ]


class KpartxStepRunner(vmdb.StepRunnerInterface):

    def get_required_keys(self):
        return ['kpartx']

    def run(self, step, settings, state):
        device = step['kpartx']
        tags = state.tags.get_tags()
        devs = self.kpartx(device)
        for tag, dev in zip(tags, devs):
            vmdb.progress('remembering {} as {}'.format(dev, tag))
            state.tags.set_dev(tag, dev)

    def kpartx(self, device):
        output = vmdb.runcmd(['kpartx', '-asv', device]).decode('UTF-8')
        for line in output.splitlines():
            words = line.split()
            if words[0] == 'add':
                name = words[2]
                yield '/dev/mapper/{}'.format(name)

    def teardown(self, step, settings, state):
        device = step['kpartx']
        vmdb.runcmd(['kpartx', '-dsv', device])



# Copyright 2018  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


import unittest


import vmdb


class TagsTests(unittest.TestCase):

    def test_lists_not_tags_initally(self):
        tags = vmdb.Tags()
        self.assertEqual(tags.get_tags(), [])

    def test_tells_if_tag_exists(self):
        tags = vmdb.Tags()
        self.assertFalse(tags.has_tag('foo'))
        tags.append('foo')
        self.assertTrue(tags.has_tag('foo'))
        self.assertEqual(tags.get_tags(), ['foo'])

    def test_remembers_order(self):
        tags = vmdb.Tags()
        tags.append('foo')
        tags.append('bar')
        self.assertTrue(tags.get_tags(), ['foo', 'bar'])

    def test_get_dev_raises_error_for_unknown_tag(self):
        tags = vmdb.Tags()
        with self.assertRaises(vmdb.UnknownTag):
            tags.get_dev('does-not-exist')

    def test_get_mount_point_raises_error_for_unknown_tag(self):
        tags = vmdb.Tags()
        with self.assertRaises(vmdb.UnknownTag):
            tags.get_mount_point('does-not-exist')

    def test_raises_error_for_reused_tag(self):
        tags = vmdb.Tags()
        tags.append('tag')
        with self.assertRaises(vmdb.TagInUse):
            tags.append('tag')

    def test_sets_dev(self):
        tags = vmdb.Tags()
        tags.append('first')
        tags.set_dev('first', '/dev/foo')
        self.assertEqual(tags.get_tags(), ['first'])
        self.assertEqual(tags.get_dev('first'), '/dev/foo')
        self.assertEqual(tags.get_mount_point('first'), None)

    def test_adds_mount_point(self):
        tags = vmdb.Tags()
        tags.append('first')
        tags.set_mount_point('first', '/mnt/foo')
        self.assertEqual(tags.get_tags(), ['first'])
        self.assertEqual(tags.get_dev('first'), None)
        self.assertEqual(tags.get_mount_point('first'), '/mnt/foo')

    def test_mount_point_is_uncached_by_default(self):
        tags = vmdb.Tags()
        tags.append('first')
        tags.set_mount_point('first', '/mnt/foo')
        self.assertFalse(tags.is_cached('first'))

    def test_mount_point_can_be_made_cached(self):
        tags = vmdb.Tags()
        tags.append('first')
        tags.set_mount_point('first', '/mnt/foo', cached=True)
        self.assertTrue(tags.is_cached('first'))

    def test_set_dev_raises_error_for_unknown_tag(self):
        tags = vmdb.Tags()
        with self.assertRaises(vmdb.UnknownTag):
            tags.set_dev('first', '/mnt/foo')

    def test_set_mount_point_raises_error_for_unknown_tag(self):
        tags = vmdb.Tags()
        with self.assertRaises(vmdb.UnknownTag):
            tags.set_mount_point('first', '/mnt/foo')

    def test_set_mount_point_raises_error_for_double_mount(self):
        tags = vmdb.Tags()
        tags.append('first')
        tags.set_mount_point('first', '/mnt/foo')
        with self.assertRaises(vmdb.AlreadyMounted):
            tags.set_mount_point('first', '/mnt/foo')

    def test_set_dev_raises_error_for_double_dev(self):
        tags = vmdb.Tags()
        tags.append('first')
        tags.set_dev('first', '/dev/foo')
        with self.assertRaises(vmdb.AlreadyHasDev):
            tags.set_dev('first', '/dev/foo')



# Copyright 2018  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


import unittest


import vmdb


class UnmountTests(unittest.TestCase):

    def setUp(self):
        self.mounts = ProcMounts()

    def unmount(self, what):
        vmdb.unmount(
            what,
            mounts=str(self.mounts),
            real_unmount=self.mounts.unmount)

    def test_raises_error_if_not_mounted(self):
        with self.assertRaises(vmdb.NotMounted):
            self.unmount('/foo')

    def test_unmounts_mounted_dir(self):
        self.mounts.mount('/dev/foo', '/foo')
        self.unmount('/foo')
        self.assertFalse(self.mounts.is_mounted('/foo'))

    def test_unmounts_mounted_dir_with_submounts(self):
        self.mounts.mount('/dev/foo', '/foo')
        self.mounts.mount('/dev/bar', '/foo/bar')
        self.unmount('/foo')
        self.assertFalse(self.mounts.is_mounted('/foo'))
        self.assertFalse(self.mounts.is_mounted('/foo/bar'))


class ProcMounts:

    def __init__(self):
        self.mounts = []

    def is_mounted(self, what):
        return any(what in mount for mount in self.mounts)

    def mount(self, device, point):
        self.mounts.append((device, point))

    def unmount(self, what):
        self.mounts = [
            mount
            for mount in self.mounts
            if what not in mount
        ]

    def __str__(self):
        return ''.join(
            '{}\n'.format(self.mount_line(mount))
            for mount in self.mounts
        )

    def mount_line(self, mount):
        return '{} {} fstype options 0 0'.format(mount[0], mount[1])
